//
// Positive (version 1.0.0.0)
//


#define _WIN32_WINNT 0x0500

#include <windows.h>
#include <tchar.h>

#define IDHK_START 1
#define IDHK_RESET 2


// Prototype declaration
BOOL InitApplication(HINSTANCE hInst, LPCTSTR className);
LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wp, LPARAM lp);
void LoadIniFile(int firstTime);
void RegHotKey(int id, TCHAR *key);
DWORD WINAPI CheckIniFileThreadProc(LPVOID d);
void WmDestroy(void);
void WmPaint(void);
void WmUser(ATOM astro);
void WmTimer(void);
void WmHotKey(int id);
void WmShowWindow(BOOL show);
void WmMouseMove(void);
void SetWindow(void);


// Global variables
int Time = 10, Position = 3, Width = 3, Height = 12, Color = 0xFF0000;
int ElapsedTime, On, OnTimeTickCount, Moving;
HWND HWnd;
POINT RDownPos;
HANDLE ThreadHandle;
BOOL reserveClose;


// Main function
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPI, LPSTR cmdLine, int cmdShow) {
	LPCTSTR className = _T("Positive");
	HWND hWnd, handle;
	MSG msg;

	// Close when the second instance executed
	CreateMutex(NULL, FALSE, _T("POSITIVE1.0.0.0"));
	if(GetLastError() == ERROR_ALREADY_EXISTS) {
		handle = FindWindow(className, NULL);
		if(handle) SendMessage(handle, WM_CLOSE, 0, 0);
		return 0;
	}
	if(!InitApplication(hInst, className)) return 0;

	hWnd = CreateWindowEx(WS_EX_TOOLWINDOW | WS_EX_TOPMOST | WS_EX_LAYERED, className,
		_T(""), WS_POPUP, 0, 0, 0, 0, NULL, NULL, hInst, NULL);
	if(!hWnd) return 0;

	while(GetMessage(&msg, NULL, 0, 0) > 0) {  // WM_QUIT or error
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.wParam;
}


// Registration of window class
BOOL InitApplication(HINSTANCE hInst, LPCTSTR className) {
	WNDCLASS wc;

	wc.style         = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;  // Class style
	wc.lpfnWndProc   = (WNDPROC)WndProc;  // Window procedure name
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = 0;
	wc.hInstance     = hInst;  // Current instance
	wc.hIcon         = NULL;
	wc.hCursor       = LoadCursor(NULL, IDC_ARROW);  // Mouse cursor (normal)
	wc.hbrBackground = NULL;
	wc.lpszMenuName  = NULL;
	wc.lpszClassName = className;

	return RegisterClass(&wc);
}


// Window procedure
LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wp, LPARAM lp) {
	DWORD thread_id;

	switch(msg) {
	case WM_CREATE:
		HWnd = hWnd;
		SetTimer(hWnd, 1, 500, NULL);
		LoadIniFile(TRUE);
		ThreadHandle = CreateThread(NULL, 0, CheckIniFileThreadProc, NULL, 0, &thread_id);
		SetWindow();
		break;
	case WM_PAINT:			WmPaint(); break;
	case WM_SHOWWINDOW:		WmShowWindow((BOOL)wp); break;
	case WM_TIMER:			WmTimer(); break;
	case WM_HOTKEY:			WmHotKey(wp); break;
	case WM_RBUTTONDBLCLK:  reserveClose = TRUE; break;
	case WM_RBUTTONUP:		if(reserveClose) SendMessage(HWnd, WM_CLOSE, 0, 0);  break;
	case WM_LBUTTONDBLCLK:
		On = !On;
		if(On) OnTimeTickCount = GetTickCount();
		break;
	case WM_LBUTTONDOWN:
		RDownPos.x = LOWORD(lp), RDownPos.y = HIWORD(lp);
		SetCapture(hWnd);
		Moving = TRUE;
		break;
	case WM_MOUSEMOVE:		WmMouseMove(); break;
	case WM_LBUTTONUP:
		ReleaseCapture();
		Moving = FALSE;
		break;
	case WM_USER:
		On = FALSE;
		LoadIniFile(FALSE);
		ElapsedTime = (ElapsedTime > Time) ? Time : ElapsedTime;
		InvalidateRect(hWnd, NULL, FALSE);
		SetWindow();
		break;
	case WM_DESTROY:        WmDestroy(); break;
	default:                return DefWindowProc(hWnd, msg, wp, lp);
	}
	return 0;
}


// Loading configuration file
void LoadIniFile(int firstTime) {
	int temp;
	TCHAR iniPath[MAX_PATH], key[3];
	UINT flag;

	GetModuleFileName(NULL, iniPath, MAX_PATH - 1);
	iniPath[lstrlen(iniPath) - 3] = _T('\0');
	lstrcat(iniPath, _T("ini"));

	// Measurement time (minutes)
	Time = GetPrivateProfileInt(_T("Timer"), _T("Time"), Time, iniPath) * 60000;
	// Size of indicator
	Width  = GetPrivateProfileInt(_T("Indicator"), _T("Width"), Width, iniPath);
	Height = GetPrivateProfileInt(_T("Indicator"), _T("Height"), Height, iniPath);
	// Position of indicator
	Position = GetPrivateProfileInt(_T("Indicator"), _T("Position"), Position, iniPath);
	if(Position == 0 || Position == 2) {
		temp = Width;
		Width = Height;
		Height = temp;
	}
	// Color of indicator
	Color = GetPrivateProfileInt(_T("Indicator"), _T("Color"), Color, iniPath);
	// Transparency of indicator
	temp = GetPrivateProfileInt(_T("Indicator"), _T("Trans"), 0, iniPath);
	SetLayeredWindowAttributes(HWnd, 0, temp, LWA_ALPHA);

	if(!firstTime) {
		UnregisterHotKey(HWnd, IDHK_START);
		UnregisterHotKey(HWnd, IDHK_RESET);
	}
	// Registration of hot keys
	GetPrivateProfileString(_T("HotKey"), _T("Start"), _T("0S"), key, 3, iniPath);
	RegHotKey(IDHK_START, key);
	GetPrivateProfileString(_T("HotKey"), _T("Reset"), _T("0R"), key, 3, iniPath);
	RegHotKey(IDHK_RESET, key);
}


// Registration of hot key
void RegHotKey(int id, TCHAR *key) {
	UINT flag;

	switch(key[0] - _T('0')) {
	case 0: flag = MOD_CONTROL | MOD_SHIFT; break;
	case 1: flag = MOD_CONTROL | MOD_ALT; break;
	case 2: flag = MOD_ALT | MOD_SHIFT; break;
	default: flag = MOD_CONTROL | MOD_SHIFT; break;
	}
	RegisterHotKey(HWnd, id, flag, (UINT)key[1]);
}


// Thread for monitoring update of configuration file
DWORD WINAPI CheckIniFileThreadProc(LPVOID d) {
	int i, pos = -1;
	TCHAR path[MAX_PATH];
	HANDLE handle;

	GetModuleFileName(NULL, path, MAX_PATH - 1);
	// Get parent folder path
	for(i = 0; path[i] != _T('\0'); ++i) {
#ifndef UNICODE
		if(IsDBCSLeadByte(path[i])) {  // for SJIS (a Japanese charactor code)
			++i;
			continue;
		}
#endif
		if(path[i] == '\\') pos = i;
	}
	// For drive root like "C:\"
	if(pos == 2) ++pos;
	path[pos] = _T('\0');

	// For getting notification when the file is written
	handle = FindFirstChangeNotification(path, FALSE, FILE_NOTIFY_CHANGE_LAST_WRITE);
	if(handle == INVALID_HANDLE_VALUE) return -1;

	// Wait until updated
	while(1) {
		if(WaitForSingleObject(handle, INFINITE) == WAIT_OBJECT_0) {
			SendMessage(HWnd, WM_USER, 0, 0);  // Notify the update
		} else {
			break;
		}
		// Set notification again
		if(!FindNextChangeNotification(handle)) break;
	}
	FindCloseChangeNotification(handle);
	return 0;
}


// Event handler for WM_DESTROY
void WmDestroy() {
	CloseHandle(ThreadHandle);
	UnregisterHotKey(HWnd, IDHK_START);
	UnregisterHotKey(HWnd, IDHK_RESET);
	PostQuitMessage(0);
}


// Event handler for WM_PAINT
void WmPaint() {
	RECT rc;
	HDC hdc;
	HBRUSH hb;
	PAINTSTRUCT paint;

	GetClientRect(HWnd, &rc);
	hdc = BeginPaint(HWnd, &paint);
	hb = CreateSolidBrush(Color);
	FillRect(hdc, &rc, hb);
	DeleteObject(hb);
	EndPaint(HWnd, &paint);
}


// Event handler for WM_TIMER
void WmTimer() {
	if(!On) return;
	ElapsedTime += GetTickCount() - OnTimeTickCount;
	OnTimeTickCount = GetTickCount();
	if(ElapsedTime >= Time) {
		ElapsedTime = Time;
		On = FALSE;
	}
	SetWindow();
}


// Event handler for WM_HOTKEY
void WmHotKey(int id) {

	switch(id) {
	case IDHK_START:
		On = !On;
		if(!On) break;
		OnTimeTickCount = GetTickCount();
		break;
	case IDHK_RESET:
		ElapsedTime = 0;
		break;
	}
	SetWindow();
}


// Event handler for WM_MOUSEMOVE
void WmMouseMove() {
	int r, b, alt;
	POINT pt;

	if(!Moving) return;

	GetCursorPos(&pt);
	pt.x -= RDownPos.x;
	pt.y -= RDownPos.y;
	alt = (GetAsyncKeyState(VK_MENU) & 0x8000);
	r = GetSystemMetrics(SM_CXSCREEN);
	b = GetSystemMetrics(SM_CYSCREEN);

	if(Position % 2) {
		ElapsedTime = pt.y * Time / (b - Height);
	} else {
		ElapsedTime = pt.x * Time / (r - Width);
	}
	if(alt) ElapsedTime = (ElapsedTime + 30000) / 60000 * 60000;
	if(ElapsedTime > Time) ElapsedTime = Time;
	SetWindow();
}


// Event handler for WM_SHOWWINDOW
void WmShowWindow(BOOL show) {
	int targetID, foregID;
	DWORD t;

	foregID  = GetWindowThreadProcessId(GetForegroundWindow(), NULL);
	targetID = GetWindowThreadProcessId(HWnd, NULL);
	AttachThreadInput(targetID, foregID, TRUE);
	SystemParametersInfo(SPI_GETFOREGROUNDLOCKTIMEOUT, 0, &t, 0);
	SystemParametersInfo(SPI_SETFOREGROUNDLOCKTIMEOUT, 0, NULL, 0);
	SetForegroundWindow(HWnd);
	SystemParametersInfo(0x2001, t, NULL, 0);
	AttachThreadInput(targetID, foregID, FALSE);
}


// Setting window size and position
void SetWindow() {
	int r, b, x, y;

	r = GetSystemMetrics(SM_CXSCREEN);
	b = GetSystemMetrics(SM_CYSCREEN);
	if(Position % 2) {
		x = (Position == 3) ? 0 : r - Width;
		y = (b - Height) * ElapsedTime / Time;
	} else {
		x = (r - Width) * ElapsedTime / Time;
		y = (Position == 0) ? 0 : b - Height;
	}
	MoveWindow(HWnd, x, y, Width, Height, TRUE);
	ShowWindow(HWnd, SW_SHOWNOACTIVATE);
}
